import glob
import os
import sys
from collections import (
    defaultdict,
    namedtuple,
)
from contextlib import (
    contextmanager,
)
from itertools import (
    chain,
    islice,
)
from typing import (
    Any,
    Dict,
    Iterable,
)

from django.conf import (
    settings,
)
from django.db import (
    close_old_connections,
    connections,
)
from django.db.utils import (
    load_backend,
)

from factory_bo.consts import (
    TAB_STR,
)


Results = namedtuple('Results', ['sorted', 'cyclic'])


@contextmanager
def substitute_default_db_connection(
    db_connection_settings: Dict[str, Any],
):
    """
    Менеджер контекста предназначен для подмены дефолтного подключения БД на
    пользовательский на время выполения сценария
    """
    temp_db_connection = connections['default']
    temp_db_connection_settings = temp_db_connection.settings_dict.copy()
    try:
        close_old_connections()
        backend = load_backend(db_connection_settings['ENGINE'])
        db_connection = backend.DatabaseWrapper(
            db_connection_settings,
            alias='default',
        )
        db_connection.connect()
        connections['default'] = db_connection

        yield

    except Exception as e:
        print(e)
    finally:
        close_old_connections()

        backend = load_backend(temp_db_connection_settings['ENGINE'])
        connections['default'] = backend.DatabaseWrapper(
            temp_db_connection_settings,
            alias='default',
        )
        connections['default'].connect()


@contextmanager
def substitute_model_get_queryset_method(model):
    """
    Контекстный менеджер подмены метода get_queryset
    стандартного менеджера модели objects, на метод базового менеджера.

    Подмена нужна для случаев когда менеджер модели по-умолчанию дополнительно фильтрует QS.
    В частности чтобы исключить фильтрацию по end_id у моделей наследованных
    от BaseModelWithEnterprise (менеджер EnterpriseDictionaryRecordManager).
    """
    try:
        has_get_queryset_back = hasattr(
            model.objects,
            '_get_queryset_back'
        )
    except RecursionError:
        # Вход в рекурсию происходит из-за метода
        # salary.core.base.models.PassThroughManagerMixin.__getattr__
        # который при попытке получить несуществующий атрибут
        # начинает рекурсивно вызывать сам себя
        # TODO BOZIK-28931
        has_get_queryset_back = False

    if not has_get_queryset_back:
        model.objects._get_queryset_back = model.objects.get_queryset

    model.objects.get_queryset = model._base_manager.get_queryset

    try:
        yield
    finally:
        model.objects.get_queryset = model.objects._get_queryset_back


class FactoryRepr(object):
    """
    Представление записи фабрики в виде строки
    """
    def __init__(
        self,
        factory_class_name: str,
        fields: Dict[str, str],
        is_formatted: bool = False,
        start_indent: str = '',
    ):
        self._factory_class_name = factory_class_name
        self._fields = fields
        self._is_formatted = is_formatted
        self._start_indent = start_indent

    def as_dict(self):
        return f'{self._factory_class_name}(**{self._fields})'

    def as_str(self):
        field_indent = (
            f'{self._start_indent}{TAB_STR}' if
            self._is_formatted else
            ''
        )
        fields_values = [
            f'{field_indent}{x}={y}'
            for x, y in self._fields.items()
        ]

        if self._is_formatted:
            fields_values_str = ',\n'.join(fields_values)
            if fields_values_str:
                fields_values_str = f'{fields_values_str},'

            repr_ = (
                f'{self._factory_class_name}(\n'
                f'{fields_values_str}\n{TAB_STR})'
            )
        else:
            fields_values_str = ', '.join(fields_values)
            repr_ = f'{self._factory_class_name}({fields_values_str})'

        return repr_


def is_pseudo_fk_field(
    field_name: str,
) -> bool:
    """
    Является ли поле псевдо внешним ключом
    """
    return (
        field_name in settings.FACTORY_BO__PSEUDO_SELF_FK_IDS or
        field_name in settings.FACTORY_BO__PSEUDO_FK_IDS
    )


def make_chunks(
    iterable: Iterable,
    size: int,
    is_list: bool = False,
):
    """
    Эффективный метод нарезки итерабельного объекта на куски
    """
    iterator = iter(iterable)

    for first in iterator:
        yield (
            list(chain([first], islice(iterator, size - 1))) if
            is_list else
            chain([first], islice(iterator, size - 1))
        )


def get_pk_slug(
    pk: str,
):
    """
    Метод преобразования первичного ключа к виду для использования, как части
    названия переменной
    """
    return pk.replace('\'', '').replace('-', '_')


def colored_stdout_output(
    message: str,
    color: int,
):
    """
    Вывод сообщения в основной поток вывода с форматированием цветом

    :param message: сообщение
    :param color: код цвета
    :return:
    """
    new_message = f'\33[{color};1m{message}\33[0m\n'

    sys.stdout.write(new_message)


def get_last_edited_fixtures():
    """
    Получение абсолютных путей последних отредактированных фикстур
    """
    fixtures_dir_path = settings.FACTORY_BO__FIXTURES_DIR_PATH

    fixtures = glob.glob(os.path.join(fixtures_dir_path, '*.json'))
    sorted_fixtures = sorted(
        fixtures,
        key=os.path.getctime,
        reverse=True,
    )

    return sorted_fixtures[1], sorted_fixtures[0]


def check_fixture_files(
    begin_file_name: str,
    end_file_name: str,
):
    """
    Проверка существования файлов. Возвращает полные пути файлов фикстур
    до и после выполнения действий предыстории
    """
    if not begin_file_name:
        raise SystemExit('Check begin_file_name parameter')

    if not end_file_name:
        raise SystemExit('Check end_file_name parameter')

    fixtures_dir_path = settings.FACTORY_BO__FIXTURES_DIR_PATH

    begin_file_path = os.path.join(fixtures_dir_path, begin_file_name)
    if not os.path.exists(begin_file_path):
        raise SystemExit('Begin file does not exists')

    end_file_path = os.path.join(fixtures_dir_path, end_file_name)
    if not os.path.exists(end_file_path):
        raise SystemExit('End file does not exists')

    return begin_file_path, end_file_path


def topological_sort(dependency_pairs):
    """
    Топологическая сортировка графов. На вход подаются ребра графа в виде
    кортежей названий вершин.

    Алгоритм отлично подходит для сортировки порядка следования таблиц
    относительно зависимостей между ними.

    django.utils.topological_sort.stable_topological_sort - работает только с
    одним графом, тут можно подать несколько графов на вход

    print( topological_sort('aa'.split()) )
    print( topological_sort('ah bg cf ch di ed fb fg hd he ib'.split()) )

    Данный код является заимствованным из открытых источников.
    Спасибо Raymond Hettinger
    """
    num_heads = defaultdict(int)
    tails = defaultdict(list)
    heads = []
    for h, t in dependency_pairs:
        num_heads[t] += 1
        if h in tails:
            tails[h].append(t)
        else:
            tails[h] = [t]
            heads.append(h)

    ordered = [h for h in heads if h not in num_heads]
    for h in ordered:
        for t in tails[h]:
            num_heads[t] -= 1
            if not num_heads[t]:
                ordered.append(t)
    cyclic = [n for n, heads in num_heads.items() if heads]

    return Results(ordered, cyclic)



from typing import (
    Set,
)


class PreparingModelRecordTagEnum:
    GETTING = 1
    CREATING = 2
    UPDATING = 3
    DELETING = 4

    values = {
        GETTING: 'Получаемая запись',
        CREATING: 'Создаваемая запись',
        UPDATING: 'Обновляемая запись',
        DELETING: 'Удаляемая запись',
    }

    ALL = (
        GETTING,
        CREATING,
        UPDATING,
        DELETING,
    )

    # Несовместимые состояния записи модели
    INCOMPATIBLES = (
        CREATING,
        UPDATING,
        DELETING,
    )

    @classmethod
    def is_incompatible(
        cls,
        tag: int,
    ):
        """
        Относится к несовместимым
        """
        return tag in cls.INCOMPATIBLES

    @classmethod
    def have_incompatible(
        cls,
        tags: Set[int],
    ):
        """
        Показывает, что набор тегов уже содержит несовместимые состояния записи
        модели
        """
        return bool(tags.intersection(cls.INCOMPATIBLES))

    @classmethod
    def is_updating(
        cls,
        tag: int,
    ):
        """
        Является обновляемым
        """
        return tag == cls.UPDATING

    @classmethod
    def is_deleting(
        cls,
        tag: int,
    ):
        """
        Является удаляемым
        """
        return tag == cls.DELETING


class DiffFormatEnum:
    PYTHON = 'py'
    SQL = 'sql'

    values = {
        PYTHON: 'Выгрузка в виде Python-файла содержащего фабрики',
        SQL: (
            'Выгрузка в виде SQL-файла с запросами сгенерированными после '
            'исполнения кода из Python-файла с фабриками'
        ),
    }


class UsingLibraryEnum:
    DATETIME = 'datetime'
    DECIMAL = 'Decimal'
    JSON = 'json.'

    values = {
        DATETIME: 'import datetime',
        DECIMAL: 'from decimal import Decimal',
        JSON: 'import json',
    }


class FixtureDestinationEnum:
    STDOUT = 'stdout'
    FILE = 'file'

    values = {
        STDOUT: 'Вывод производится в стандартный поток вывода',
        FILE: 'Вывод производится в файл',
    }


class EtalonFixtureGeneratorModeEnum:
    MERGE = 'merge'
    REPLACE = 'replace'

    values = {
        MERGE: (
            'Режим объединения данных существующих фикстур с данными уже '
            'существующих'
        ),
        REPLACE: 'Режим замены без учета уже существующих данных',
    }


class FactoryUseTypeEnum:
    DEFAULT = 1
    CUSTOM = 2

    values = {
        DEFAULT: 'Фабрика используется по умолчанию',
        CUSTOM: (
            'Фабрику нужно явно указать в словаре соответствия model_name, '
            'factory_class_name декоратора теста'
        ),
    }

    @classmethod
    def is_default(
        cls,
        type_: int,
    ):
        """
        Фабрика является дефолтной для модели
        """
        return type_ == cls.DEFAULT

    @classmethod
    def is_custom(
        cls,
        type_: int,
    ):
        """
        Фабрика является кастомной для модели
        """
        return type_ == cls.CUSTOM

    @classmethod
    def exists(
        cls,
        use_type: int
    ):
        """
        Проверяется существование типа использования
        """
        return use_type in cls.values.keys()



from setuptools import (
    find_packages,
    setup,
)

VERSION = '0.0.2'


try:
    long_description = open('README.md', 'rt').read()
except IOError:
    long_description = ''

with open('requirements.txt', encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [
    x.strip()
    for x in all_reqs if
    'git+' not in x
]

setup(
    name='factory_bo',
    version=VERSION,

    description='Electronic Document Management',
    long_description=long_description,

    author='Stepan Lushchiy',
    author_email='s.lushchiy@bars.group',

    url='http://docs.budg.bars.group/factory_bo/',
    download_url='',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
    ],

    platforms=['Any'],

    scripts=[],

    provides=[],

    namespace_packages=[],
    packages=find_packages(),
    include_package_data=True,

    package_data={
        '': [],
    },

    install_requires=install_requires,

    zip_safe=False,
)



from factory_bo.consts import (
    TAB_STR,
)


WRONG_FIXTURE_FILE_EXTENSION_ERROR = (
    'You should send absolute path of fixture file with .json extension, '
    'but send {destination_file_path}'
)

SEVERAL_MODEL_DAFAULT_FACTORIES_ERROR = (
    'Model "{model_label}" has several default factories: {factories}!'
)

WRONG_FACTORY_USE_TYPE_ERROR = (
    'Wrong factory use type "{factory_use_type}" of {factory_class_name}!'
)

PYTHON_FIXTURE_PATH_HAVE_NOT_PY_EXTENSION_ERROR = (
    'Python fuxture path have not .py extension - "{fixture_path}"!'
)

FACTORIES_FOR_MODEL_NOT_FOUND = (
    'Factories for model "{model_label}" not found.'
)

MODEL_WITHOUT_DEFAULT_FACTORY_ERROR = (
    'Models - ({model_labels}) without default factories!'
)

FACTORY_CLASS_FOR_MODEL_NOT_FOUND = (
    'Factory "{factory_class_name}" for model "{model_label}" not found!'
)

FACTORY_BO__FIXTURES_DIR_PATH_NOT_FOUND = (
    'Please set FIXTURES_DIR_PATH in the [factory_bo] section of'
    ' configuration file'
)

WRONG_DIFF_FORMAT = (
    'Wrong diff_format - "{}"! You can choose only py or sql formats.'
)

DISALLOWED_DJANGO_GET_OR_CREATE_FIELDS_FOUND_ERROR = (
    'Found dissallowed django_get_of_create fields - ({disallowed_fields}) of '
    'factory class "{factory_class}"!'
)

IMPORT_DEFAULT_MANAGER_FACTORY_STR = (
    f'from runner.factories.base import (\n{TAB_STR}DefaultManagerFactory,\n)\n'
)

NON_UNIQUE_DJANGO_GET_OR_CREATE_RECORD_FOUND_ERROR = (
    'Non unique django get or create record found error! Factory - '
    '{factory_class_name}, key - {key}'
)

FACTORY_CLASS_STR = (
    f'\n\n'
    f'class {{factory_class_name}}({{base_factory_class}}):\n'
    f'\n'
    f'{TAB_STR}class Meta:\n'
    f'{TAB_STR*2}model = \'{{model_label}}\'\n'
    f'{TAB_STR*2}# django_get_or_create = (\n'
    f'{{django_get_or_create_str}}\n'
    f'{TAB_STR*2}# )'
    f'\n'
)

FACTORY_CLASS_WITHOUT_COMMENT_STR = (
    f'\n\n'
    f'class {{factory_class_name}}({{base_factory_class}}):\n'
    f'\n'
    f'{TAB_STR}class Meta:\n'
    f'{TAB_STR*2}model = \'{{model_label}}\'\n'
    f'{TAB_STR*2}django_get_or_create = (\n'
    f'{{django_get_or_create_str}}\n'
    f'{TAB_STR*2})'
    f'\n'
)

IMPORT_FACTORY_USE_TYPE_ENUM_STR = (
    '\nfrom runner.factories.enums import FactoryUseTypeEnum'
)

USING_GENERATED_FACTORY_WARNING = (
    'Attention! You are use generated factory - "{generated_factory}"! Please '
    'actualize factory and remove "Generated" prefix.'
)

RECORD_FOR_FK_FIELD_NOT_FOUND_ERROR = (
    'For fk field "{field_name}" in model "{model_label}" related record '
    '"{related_record}" with id={field_value_id} in storage not found.'
    'Probably, problem is in hard binding to id in the old fixture.'
)



from django.apps import (
    apps,
)
from django.core.management import (
    BaseCommand,
)
from django.db.models import (
    ForeignKey,
)


class Command(BaseCommand):
    """
    Команда для поиска циклических зависимостей на уровне таблиц
    """
    def handle(self, *args, **options):
        models = apps.get_models(include_auto_created=True)
        model_relation_pairs = set()

        for model in models:
            for field in model._meta.fields:
                if (
                    isinstance(field, ForeignKey) and
                    model._meta.label != field.related_model._meta.label
                ):
                    model_relation_pairs.add(
                        (
                            model._meta.label,
                            field.related_model._meta.label
                        )
                    )

        for model1, model2 in model_relation_pairs:
            if (model2, model1) in model_relation_pairs:
                print(f'{model2},{model1}')
